# vim: fdm=indent
#author:     Fabio Zanini
#date:       12/08/19
#content:    Fetch cell atlas data, averaged by cell type
__all__ = ['AtlasFetcher']

import os
import numpy as np
import pandas as pd
import requests
import io
import loompy
from anndata import AnnData
import tempfile


class AtlasFetcher(object):
    '''Fetch cell atlas data from https://iosonofabio.github.io/atlas_landmarks/'''
    atlas_table = None

    def fetch_atlas_table(self):
        '''Fetch atlas table from GitHub repo'''
        url = 'https://github.com/iosonofabio/atlas_landmarks/raw/master/table.tsv'
        r = requests.get(url)
        table = pd.read_csv(io.BytesIO(r.content), sep='\t', index_col=0)

        self.atlas_table = table

    def list_atlases(self):
        '''List atlases available on GitHub repo'''
        if self.atlas_table is None:
            self.fetch_atlas_table()
        return self.atlas_table.copy()

    def fetch_atlas(self, atlas_name, kind='average', cell_types=None):
        '''Fetch an atlas from https://iosonofabio.github.io/atlas_landmarks/

        Args:
            atlas_name (str): the name of the atlas (see atlas table)
            kind (str): must be 'average' for the average expression of each
             cell type, or 'subsample' for a subsample of the atlas.
            cell_types (None or list of str): restrict to certain cell types
             within the atlas. None (default) means all cell types will be
             retained.

        Returns:
            a dictionary with two keys. 'counts' contains the gene expression
            count table. For kind == 'average', 'number_of_cells' include the
            number of cells within each cell type. For kind == 'subsample',
            'cell_types' includes the cell types of the subsample.
        '''
        if kind not in ('average', 'subsample'):
            raise ValueError('kind must be one of "average" and "subsample"')

        if self.atlas_table is None:
            self.fetch_atlas_table()

        if kind == 'subsample':
            url = self.atlas_table.at[atlas_name, 'URL_subsample']
        else:
            url = self.atlas_table.at[atlas_name, 'URL_average']

        r = requests.get(url)

        # Use a temp file, loompy has its own quirks
        fd, path = tempfile.mkstemp()
        try:
            with os.fdopen(fd, 'wb') as tmp:
                # do stuff with temp file
                tmp.write(r.content)

            with loompy.connect(path) as dsl:
                matrix = dsl.layers[''][:, :]
                features = dsl.ra['GeneName']
                cell_types_all = dsl.ca['CellType']
                if kind == 'average':
                    n_of_cells = dsl.ca['NumberOfCells']
                else:
                    cell_names = dsl.ca['CellName']

        finally:
            os.remove(path)

        # Package into anndata
        if kind == 'subsample':
            meta = pd.Series(
                data=cell_types_all,
                index=cell_names,
                )

            if cell_types is not None:
                ind_ct = meta.isin(cell_types).values.nonzero()[0]
                meta = meta.loc[ind_ct]
                matrix = matrix[:, ind_ct]

            meta.name = 'CellType'
            obs = meta.to_frame()
            obs['CellID'] = obs.index
            var = pd.Series(features, index=features)
            var.name = 'GeneName'
            var = var.to_frame()
            adata = AnnData(
                X=matrix.T,
                obs=obs,
                var=var,
                )
            adata.obs_names = adata.obs['CellID']
            adata.var_names = adata.var['GeneName']

        else:
            number_of_cells = pd.Series(
                data=n_of_cells,
                index=cell_types_all,
                )

            if cell_types is not None:
                ind_ct = number_of_cells.index.isin(cell_types).values.nonzero()[0]
                number_of_cells = number_of_cells[ind_ct]
                matrix = matrix[:, ind_ct]

            number_of_cells.name = 'NumberOfCells'
            obs = number_of_cells.to_frame()
            obs['CellID'] = obs.index
            var = pd.Series(features, index=features)
            var.name = 'GeneName'
            var = var.to_frame()
            adata = AnnData(
                X=matrix.T,
                obs=obs,
                var=var,
                )
            adata.obs_names = adata.obs['CellID']
            adata.var_names = adata.var['GeneName']

        return adata

    def fetch_multiple_atlases(
            self,
            atlas_names,
            kind='average',
            join='keep_first',
            ):
        '''Fetch and combine multiple atlases

        Args:
            atlas_names (list of str or list of dict): the names of the atlases
             (see atlas table). If a list of dict, it can be used to fetch only
             certain cell types from certain atlases. In that case, every
             element of the list must be a dict with two key-value pairs:
             'atlas_name' is the atlas name as of the atlas table, and
             'cell_types' must be a list of cell types to retain. Example:
             atlas_names=[{'atlas_name': 'Enge_2017', 'cell_tpes': ['alpha']}]
             would load the atlas Enge_2017 and only retain alpha cells.
            kind (str): must be 'average' for the average expression of each
             cell type, or 'subsample' for a subsample of the atlas.
            join (str): must be 'keep_first', 'union', or 'intersection'. This
             argument decides what to do with features that are not present
             in all atlases. 'keep_first' keeps the features in the first
             atlas and pads the other atlases with zeros, 'union' pads every
             atlas that is missing a feature and 'intersection' only keep
             features that are in all atlases.

        '''
        if kind not in ('average', 'subsample'):
            raise ValueError('kind must be one of "average" and "subsample"')

        ds = {}
        if len(atlas_names) == 0:
            return ds

        # Fetch data for all atlases
        first_atlas = None
        for atlas_name in atlas_names:
            if isinstance(atlas_name, str):
                atlas_ctypes = None
            elif isinstance(atlas_name, dict):
                atlas_ctypes = atlas_name['cell_types']
                atlas_name = atlas_name['atlas_name']
            else:
                raise ValueError('List of atlases not understood')

            if first_atlas is None:
                first_atlas = atlas_name
            ds[atlas_name] = self.fetch_atlas(
                    atlas_name,
                    kind=kind,
                    cell_types=atlas_ctypes,
                    )

        # Get features, list of all cells, etc.
        # Rename cells to ensure there are no duplicates
        cell_names = []
        cell_names_new = []
        features_list = []
        for at, d in ds.items():
            cell_names.extend(d.obs_names.values.tolist())
            cell_names_new.extend(
                    ['{:}_{:}'.format(at, x) for x in d.obs_names])
            features_list.append(d.var_names.values)
        cell_names = np.array(cell_names)
        cell_names_new = np.array(cell_names_new)

        # Deal with non-intersecting features
        if len(features_list) == 1:
            features = features_list[0]
        elif join == 'intersection':
            features_list = [set(x) for x in features_list]
            features = np.sort(set.intersection(features_list))
        elif join == 'union':
            features_list = [set(x) for x in features_list]
            features = np.sort(set.union(features_list))
        elif join == 'keep_first':
            features = np.sort(features_list[0])

        # Fill the combined dataset, including metadata
        matrix = np.empty((len(features), len(cell_names)), np.float32)
        if kind == 'average':
            n_cells_per_type = np.empty(len(cell_names), int)
        else:
            cell_types = []
        cell_dataset = []
        i = 0
        features_set = set(features)
        for at, d in ds.items():
            n = d.X.shape[0]

            # Pad missing features
            if (join == 'union') or ((join == 'keep_first') and (at != first_atlas)):
                fea_this = set(d.var_names)
                fea_missing = list(features_set - fea_this)
                fea_all = list(d.var_names) + fea_missing
                submat = np.zeros((len(fea_all), n), np.float32)
                submat[:len(fea_this)] = d.X.T

                idx = pd.Series(np.arange(len(fea_all)), index=fea_all)
                idx = idx[features].values
                submat = submat[idx]
            else:
                submat = d[:, features].X.T

            matrix[:, i: i+n] = submat

            if kind == 'average':
                n_cells_per_type[i: i+n] = d.obs['NumberOfCells'].values
            else:
                cell_types.extend(d.obs['CellType'].values.tolist())
            cell_dataset.extend([at] * d.X.shape[0])
            i += n

        cell_dataset = pd.Series(
            data=cell_dataset,
            index=cell_names_new,
            name='Dataset',
            )

        obs = cell_dataset.to_frame()
        obs['CellID'] = obs.index
        var = pd.Series(features, index=features, name='GeneName').to_frame()
        adata = AnnData(
            X=matrix.T,
            obs=obs,
            var=var,
            )
        adata.obs_names = cell_names_new
        adata.var_names = features

        if kind == 'average':
            number_of_cells = pd.Series(
                data=n_cells_per_type,
                index=cell_names_new,
                )
            adata.obs['NumberOfCells'] = number_of_cells
        else:
            cell_types = pd.Series(
                data=cell_types,
                index=cell_names_new,
                )
            adata.obs['CellType'] = cell_types

        return adata



# vim: fdm=indent
# author:     Fabio Zanini
# date:       5/08/19
# content:    Atlas subsampling
__all__ = ['subsample_atlas', 'average_atlas']

import warnings
import numpy as np
import pandas as pd
import anndata


def subsample_atlas(
        atlas,
        cell_type_column='CellType',
        n_cells=20,
        ):
    '''Subsample atlas across cell types

    Args:
        atlas (anndata.Anndata): The dataset to subsample.
        cell_type_column (str): The name of the metadata column to use for
            subsampling. For each unique value of this column, a certain number
            of cells will be samples (see `n_cells` argument).
        n_cells (int or dict): How many cells to sample from each cell type. If
            an int, the same number of cells will be sampled from all cell
            type. If a dict, the keys must be unique values of cell types and
            the values must be zero or positive ints that set the number of
            cells for each cell type. If a dict, the keys with zero value
            indicate what cell types to skip in the subsample.


        Note: If the atlas contains fewer than the requested number of cells
        for a cell type, all cell from that type will be sampled.
    '''
    cell_type = atlas.obs[cell_type_column]

    if np.isscalar(n_cells):
        n_celld = cell_type.value_counts()
        ct_unique = n_celld.index.values
        n_celld = {ct: n_cells for ct in ct_unique}
    else:
        n_celld = n_cells

    inds = []
    for ct, nc in n_celld.items():
        indi = (cell_type == ct).values.nonzero()[0]
        if nc < len(indi):
            indi = np.random.choice(indi, size=nc, replace=False)
            indi.sort()
        inds.extend(list(indi))

    subsample = atlas[inds]

    return subsample


def average_atlas(
        atlas,
        cell_type_column='CellType',
        ):
    '''Subsample atlas across cell types

    Args:
        atlas (anndata.Anndata): The dataset to subsample.
        cell_type_column (str): The name of the metadata column to use for
            subsampling. For each unique value of this column, a certain number
            of cells will be samples (see `n_cells` argument).

        Note: If the atlas contains fewer than the requested number of cells
        for a cell type, all cell from that type will be sampled.
    '''
    cell_type = atlas.obs[cell_type_column]
    n_celld = cell_type.value_counts()
    ct_unique = list(n_celld.index)

    matrix = np.zeros((len(ct_unique), atlas.shape[1]), np.float32)

    for ict, ct in enumerate(ct_unique):
        indi = (cell_type == ct).values.nonzero()[0]
        mati = atlas.X[indi]
        if not isinstance(mati, np.ndarray):
            # Make a dense matrix, it's ok since there are not many types
            mati = mati.toarray()
        matrix[ict] = mati.mean(axis=0)

    ave = anndata.AnnData(
            X=matrix,
            obs={'CellType': ct_unique, 'NumberOfCells': n_celld.values},
            )
    ave.obs_names = ct_unique
    ave.var = atlas.var

    return ave



# vim: fdm=indent
# author:     Fabio Zanini
# date:       30/08/19
# content:    Test fetching atlases from
#             https://iosonofabio.github.io/atlas_averages/
import numpy as np
import pandas as pd
from anndata import AnnData
from northstar import AtlasFetcher


def test_constructor():
    af = AtlasFetcher()


def test_fetch_table():
    af = AtlasFetcher()
    af.fetch_atlas_table()

    assert(af.atlas_table is not None)


def test_list_atlases():
    af = AtlasFetcher()
    table = af.list_atlases()
    assert(table is not None)


def test_fetch_Darmanis_2015():
    af = AtlasFetcher()
    atlas = af.fetch_atlas('Darmanis_2015')
    assert(isinstance(atlas, AnnData))
    assert('NumberOfCells' in atlas.obs.columns)


def test_fetch_Enge_2017():
    af = AtlasFetcher()
    atlas = af.fetch_atlas('Enge_2017')
    assert(isinstance(atlas, AnnData))
    assert('NumberOfCells' in atlas.obs.columns)

    counts = atlas[:, ['INS', 'GCG', 'PPY']]
    assert(counts['beta', 'INS'].X > counts['acinar', 'INS'].X)


def test_fetch_multiple():
    af = AtlasFetcher()
    atlas = af.fetch_multiple_atlases(['Darmanis_2015', 'Enge_2017'])
    assert(isinstance(atlas, AnnData))
    assert('NumberOfCells' in atlas.obs.columns)


def test_fetch_Darmanis_2015_subsample():
    af = AtlasFetcher()
    atlas = af.fetch_atlas('Darmanis_2015', kind='subsample')
    assert(isinstance(atlas, AnnData))
    assert('CellType' in atlas.obs.columns)


def test_fetch_multiple_subsample():
    af = AtlasFetcher()
    atlas = af.fetch_multiple_atlases(
            ['Darmanis_2015', 'Enge_2017'],
            kind='subsample')
    assert(isinstance(atlas, AnnData))
    assert('CellType' in atlas.obs.columns)
    assert('Dataset' in atlas.obs.columns)



# vim: fdm=indent
# author:     Fabio Zanini
# date:       17/06/19
# content:    Test the algorithm on same artificial data
import numpy as np
import pandas as pd
from northstar import AtlasFetcher, Averages, Subsample


def test_embed_averages():
    aname = 'Darmanis_2015_nofetal'
    atlas = AtlasFetcher().fetch_atlas(
            aname, kind='subsample')

    ind = [
        0, 2, 5, 8, 10, 15, 20, 25, 28, 30, 35, 38,
        40, 45, 50, 60, 70, 75, 80, 90]
    matrix = atlas[ind].copy()

    sa = Averages(
            aname,
            n_features_per_cell_type=2,
            n_features_overdispersed=5,
            n_pcs=9,
            )
    sa.fit(matrix)
    vs = sa.embed()
    assert(vs.shape[1] == 4)


def test_embed_subsample():
    aname = 'Darmanis_2015_nofetal'
    atlas = AtlasFetcher().fetch_atlas(
            aname, kind='subsample')

    ind = [
        0, 2, 5, 8, 10, 15, 20, 25, 28, 30, 35, 38,
        40, 45, 50, 60, 70, 75, 80, 90]
    matrix = atlas[ind].copy()

    sa = Subsample(
            aname,
            n_features_per_cell_type=2,
            n_features_overdispersed=5,
            n_pcs=9,
            )
    sa.fit(matrix)
    vs = sa.embed()
    assert(vs.shape[1] == 4)


if __name__ == '__main__':

    test_embed_averages()
    test_embed_subsample()



# vim: fdm=indent
# author:     Fabio Zanini
# date:       17/06/19
# content:    Atlas averages
__all__ = ['Averages']

import warnings
import numpy as np
import pandas as pd
from anndata import AnnData
import leidenalg
from .fetch_atlas import AtlasFetcher
from .cluster_with_annotations import ClusterWithAnnotations



class Averages(object):
    '''Annotate new cell types using averages of an atlas'''

    def __init__(
            self,
            atlas,
            n_cells_per_type=None,
            n_features_per_cell_type=30,
            n_features_overdispersed=500,
            features_additional=None,
            n_pcs=20,
            n_neighbors=10,
            n_neighbors_out_of_atlas=5,
            distance_metric='correlation',
            threshold_neighborhood=0.8,
            clustering_metric='cpm',
            resolution_parameter=0.001,
            normalize_counts=True,
            join='keep_first',
            ):
        '''Prepare the model for cell annotation

        Args:
            atlas (str, list of str, list of dict, dict, or AnnData): cell
             atlas to use. Generally there are two kind of choices:

             The first possibility selects the corresponding cell atlas or
             atlases from northstar's online list. The names of currently
             available dataset is here:

             https://github.com/iosonofabio/atlas_averages/blob/master/table.tsv

             (check the first column for atlas names). If a list of
             str, multiple atlases will be fetched and combined. Only features
             that are in all atlases will be kept. If you use this feature, be
             careful to not mix atlases from different species. If a list of
             dict, it merges atlases as above but you can specify what cell
             types to fetch from each atlas. Each element of the list must be a
             dict with two key-value pairs: 'atlas_name' is the atlas name, and
             'cell_types' must be a list of cell types to retain. Example:
             atlas=[{'atlas_name': 'Enge_2017', 'cell_types': ['alpha']}] would
             load the atlas Enge_2017 and only retain alpha cells. You can also
             use a dict to specify a single atlas and to retain only certain cell
             types. The format is as above, e.g. to select only alpha cells
             from Enge_2017 you can use:

             atlas={'atlas_name': 'Enge_2017', 'cell_types': ['alpha']}.

            The second possibility is to use a custom atlas (e.g. some
             unpublished data). 'atlas' must be an AnnData object with cell
             type averages ("cells") as rows and genes as columns and one cell
             metadata column 'NumberOfCells' describing the number of cells
             for each cell type. In other words:

                 adata.obs['NumberOfCells']

             must exist, and:

                adata.obs_names

            must contain the known cell types.

            n_cells_per_type (None or int): if None, use the number of cells
             per type from the atlas. Else, fix it to this number for all types.

            n_features_per_cell_type (int): number of features marking each
             fixed column (atlas cell type). The argument 'features' takes
             priority over this one.

            n_features_overdispersed (int): number of unbiased, overdispersed
             features to be picked from the new dataset. The argument
             'features' takes priority over this one.

            features_additional (list of str or None): additional features to
             keep on top of automatic selection. The argument 'features' takes
             priority over this one.

            n_pcs (int): number of principal components to keep in the weighted
             PCA.

            n_neighbors (int): number of neighbors in the similarity graph.

            n_neighbors_out_of_atlas (int): number of neighbors coming out of
             the atlas nodes into the new dataset.

            distance_metric (str): metric to use as distance. It should be a
             metric accepted by scipy.spatial.distance.cdist.

            threshold_neighborhood (float): do not consider distances larger than this as
             neighbors

            clustering_metric (str): 'cpm' (default, Cell Potts Model) or
             'modularity'. Sets the type of partition used in the clustering
             step.

            resolution_parameter (float): number between 0 and 1 that sets
             how easy it is for the clustering algorithm to make new clusters

            normalize_counts (bool): whether to renormalize the counts at the
             merging stage to make sure atlas and new data follow the same
             normalization. Be careful if you turn this off.

            join (str): must be 'keep_first', 'union', or 'intersection'. This
             argument is used when sourcing multiple atlases and decides what
             to do with features that are not present in all atlases.
             'keep_first' keeps the features in the first atlas and pads the
             other atlases with zeros, 'union' pads every atlas that is missing
             a feature and 'intersection' only keep features that are in all
             atlases.
        '''

        self.atlas = atlas
        self.n_cells_per_type = n_cells_per_type
        self.n_features_per_cell_type = n_features_per_cell_type
        self.n_features_overdispersed = n_features_overdispersed
        self.features_additional = features_additional
        self.n_pcs = n_pcs
        self.n_neighbors = n_neighbors
        self.n_neighbors_out_of_atlas = n_neighbors_out_of_atlas
        self.distance_metric = distance_metric
        self.threshold_neighborhood = threshold_neighborhood
        self.clustering_metric = clustering_metric
        self.resolution_parameter = resolution_parameter
        self.normalize_counts = normalize_counts
        self.join = join

    def fit(self, new_data):
        '''Run with averages of the atlas

        Args:
            new_data (pandas.DataFrame or anndata.AnnData): the new data to be
             clustered. If a dataframe, t must have features as rows and
             cell names as columns (as in loom files). anndata uses the opposite
             convention, so it must have cell names as rows (obs_names) and
             features as columns (var_names) and this class will transpose it.

        Returns:
            None, but this instance of Averages acquired the property
            `membership` containing the cluster memberships (cell types) of the
            columns except the first n_fixed. The first n_fixed columns are
            assumes to have distinct memberships in the range [0, n_fixed - 1].
        '''
        self.new_data = new_data

        self._check_init_arguments()
        self.fetch_atlas_if_needed()
        self.compute_feature_intersection()
        self._check_feature_intersection()
        self.prepare_feature_selection()
        self.select_features()
        self._check_feature_selection()
        self.merge_atlas_newdata()
        self.compute_pca()
        self.compute_similarity_graph()
        self.cluster_graph()

    def fit_transform(self, new_data):
        '''Run with averages of the atlas and return the cell types

        Args:
            new_data (pandas.DataFrame or anndata.AnnData): the new data to be
             clustered. If a dataframe, t must have features as rows and
             cell names as columns (as in loom files). anndata uses the opposite
             convention, so it must have cell names as rows (obs_names) and
             features as columns (var_names) and this class will transpose it.

        Returns:
            the cluster memberships (cell types) of the
            columns except the first n_fixed. The first n_fixed columns are
            assumes to have distinct memberships in the range [0, n_fixed - 1].
        '''
        self.fit(new_data)
        return self.membership

    def _check_init_arguments(self):
        # Check atlas
        at = self.atlas

        if isinstance(at, str):
            pass
        elif isinstance(at, list) or isinstance(at, tuple):
            for elem in at:
                if isinstance(elem, str):
                    pass
                elif isinstance(elem, dict):
                    if 'atlas_name' not in elem:
                        raise ValueError('List of atlases: format incorrect')
                    if 'cell_types' not in elem:
                        raise ValueError('List of atlases: format incorrect')
                else:
                    raise ValueError('List of atlases: format incorrect')

        elif isinstance(at, dict) and ('atlas_name' in at) and \
                ('cell_types' in at):
            pass

        elif isinstance(at, AnnData):
            if 'NumberOfCells' not in at.obs:
                raise AttributeError(
                        'atlas must have a "NumberOfCells" obs column')

        else:
            raise ValueError('Atlas not formatted correctly')


        # Convert new data to anndata if needed
        nd = self.new_data
        if isinstance(nd, AnnData):
            pass
        elif isinstance(nd, pd.DataFrame):
            # AnnData uses features as columns, so transpose and convert
            # (the assumption is that in the DataFrame convention, rows are
            # features)
            nd = AnnData(
                X=nd.values.T,
                obs={'CellID': nd.columns.values},
                var={'GeneName': nd.index.values},
                )
            nd.obs_names = nd.obs['CellID']
            nd.var_names = nd.var['GeneName']
            self.new_data = nd
        else:
            raise ValueError(
                'New data must be an AnnData object or pd.DataFrame',
                )

        # New data could be too small to do PCA
        n_newcells, n_newgenes = self.new_data.shape
        if n_newgenes < self.n_pcs:
            warnings.warn(
                ('The number of features in the new data is lenn than ' +
                 'the number of PCs, so northstar might give inaccurate ' +
                 'results'))

        if n_newcells < self.n_pcs:
            warnings.warn(
                ('The number of cells in the new data is lenn than ' +
                 'the number of PCs, so northstar might give inaccurate ' +
                 'results'))

        if min(n_newgenes, n_newcells) < self.n_pcs:
            warnings.warn('Reducing the number of PCs to {:}'.format(
                min(n_newgenes, n_newcells)))
            self.n_pcs = min(n_newgenes, n_newcells)

        # New data could be too small for knn
        if n_newcells < self.n_neighbors + 1:
            warnings.warn(
                ('The number of cells in the new data is less than the ' +
                 'number of neighbors requested for the knn: reducing the ' +
                 'number of graph neighbors to {:}'.format(
                     max(1, n_newcells - 1)),
                 ))
            self.n_neighbors = max(1, n_newcells - 1)

        nf1 = self.n_features_per_cell_type
        nf2 = self.n_features_overdispersed
        nf3 = self.features_additional
        if not isinstance(nf1, int):
            raise ValueError('n_features_per_cell_type must be an int >= 0')
        if not isinstance(nf1, int):
            raise ValueError('n_features_overdispersed must be an int >= 0')
        if (nf1 < 1) and (nf2 < 1) and (nf3 < 1):
            raise ValueError('No features selected')

    def _check_feature_intersection(self):
        L = len(self.features_ovl)
        if L == 0:
            raise ValueError(
                ('No overlapping features in atlas and new data, are gene ' +
                 'names correct for this species?'))
        if L < 50:
            warnings.warn(
                ('Only {:} overlapping features found in atlas and new ' +
                 'data'.format(L)))

    def _check_feature_selection(self):
        L = len(self.features)
        if L == 0:
            raise ValueError(
                ('No features survived selection, check nortstar parameters'))
        if L < self.n_pcs:
            warnings.warn(
                ('Only {0} features selected, reducing PCA to {0} components'.format(L)))
            self.n_pcs = L

    def fetch_atlas_if_needed(self):
        '''Fetch atlas(es) if needed'''
        at = self.atlas

        if isinstance(at, str):
            self.atlas = AtlasFetcher().fetch_atlas(
                    at,
                    kind='average',
                    )

        elif isinstance(self.atlas, list) or isinstance(self.atlas, tuple):
            self.atlas = AtlasFetcher().fetch_multiple_atlases(
                    at,
                    kind='average',
                    join=self.join,
                    )

        elif isinstance(at, dict) and ('atlas_name' in at) and \
                ('cell_types' in at):
            self.atlas = AtlasFetcher().fetch_atlas(
                    at['atlas_name'],
                    kind='average',
                    cell_types=at['cell_types'],
                    )

    def compute_feature_intersection(self):
        '''Calculate the intersection of features between atlas and new data'''
        # Intersect features
        self.features_atlas = self.atlas.var_names.values
        self.features_newdata = self.new_data.var_names.values
        self.features_ovl = np.intersect1d(
                self.features_atlas,
                self.features_newdata,
                )

    def prepare_feature_selection(self):
        # Cell names and types
        self.cell_types_atlas = self.atlas.obs_names
        self.cell_names_atlas = self.atlas.obs_names
        self.cell_names_newdata = self.new_data.obs_names
        ctypes_ext = []
        cnames_ext = []
        if self.n_cells_per_type is None:
            ncells_per_ct = self.atlas.obs['NumberOfCells'].astype(np.int64)
        else:
            ncells_per_ct = [self.n_cells_per_type] * self.atlas.shape[0]
        for i, ni in enumerate(ncells_per_ct):
            for ii in range(ni):
                ctypes_ext.append(self.cell_types_atlas[i])
                cnames_ext.append(self.cell_types_atlas[i]+'_{:}'.format(ii+1))
        self.cell_types_atlas_extended = ctypes_ext
        self.cell_names_atlas_extended = cnames_ext

        # Numbers
        self.n_atlas = self.atlas.shape[0]
        self.n_newdata = self.new_data.shape[0]
        self.n_total = self.n_atlas + self.n_newdata
        self.n_atlas_extended = len(self.cell_names_atlas_extended)
        self.n_total_extended = self.n_atlas_extended + self.n_newdata

        # Cell numbers
        self.sizes = np.ones(self.n_total, np.float32)
        if self.n_cells_per_type is not None:
            self.sizes[:self.n_atlas] *= self.n_cells_per_type
        else:
            self.sizes[:self.n_atlas] = self.atlas.obs['NumberOfCells'].astype(
                    np.float32)

    def select_features(self):
        '''Select features among the overlap of atlas and new data

        Returns:
            ndarray of feature names.

        '''
        features_atlas = self.features_atlas
        features_newdata = self.features_newdata
        features_ovl = list(self.features_ovl)
        features_add = self.features_additional

        n_atlas = self.n_atlas
        nf1 = self.n_features_per_cell_type
        nf2 = self.n_features_overdispersed

        features = set()

        # Atlas markers
        if (nf1 > 0) and (n_atlas > 1):
            matrix = self.atlas.X
            for icol in range(n_atlas):
                ge1 = matrix[icol]
                ge2 = (matrix.sum(axis=0) - ge1) / (n_atlas - 1)
                fold_change = np.log2(ge1 + 0.1) - np.log2(ge2 + 0.1)
                tmp = np.argsort(fold_change)[::-1]
                ind_markers_atlas = []
                for i in tmp:
                    if features_atlas[i] in features_ovl:
                        ind_markers_atlas.append(i)
                    if len(ind_markers_atlas) == nf1:
                        break
                # Add atlas markers
                features |= set(features_atlas[ind_markers_atlas])

        # Overdispersed features from new data
        if nf2 > 0:
            if nf2 >= len(features_ovl):
                features |= set(features_ovl)
            else:
                matrix = self.new_data.X
                nd_mean = matrix.mean(axis=0)
                nd_var = matrix.var(axis=0)
                fano = (nd_var + 1e-10) / (nd_mean + 1e-10)
                tmp = np.argsort(fano)[::-1]
                ind_ovd_newdata = []
                for i in tmp:
                    if features_newdata[i] in features_ovl:
                        ind_ovd_newdata.append(i)
                    if len(ind_ovd_newdata) == nf2:
                        break
                # Add overdispersed features
                features |= set(features_newdata[ind_ovd_newdata])

        # Additional features
        if features_add is not None:
            features |= (set(features_add) & set(features_ovl))

        self.features = np.array(list(features))

    def merge_atlas_newdata(self):
        '''Merge atlas data and the new data after feature selection

        NOTE: is self.normalize is True, the merged count matrix is normalized
        by 1 million total counts.
        '''
        features = self.features
        L = len(features)
        N1 = self.n_atlas
        N = self.n_total

        # This is the largest memory footprint of northstar
        matrix = np.empty((N, L), dtype=np.float32)

        # Find the feature indices for atlas
        ind_features_atlas = pd.Series(
            np.arange(len(self.features_atlas)),
            index=self.features_atlas,
            ).loc[features].values
        matrix[:N1] = self.atlas.X[:, ind_features_atlas]

        # Find the feature indices for new data
        ind_features_newdata = pd.Series(
            np.arange(len(self.features_newdata)),
            index=self.features_newdata,
            ).loc[features].values
        matrix[N1:] = self.new_data.X[:, ind_features_newdata]

        # The normalization function also sets pseudocounts
        if self.normalize_counts:
            matrix = 1e6 * (matrix.T / (matrix.sum(axis=1) + 0.1)).T

        self.matrix = matrix

    def compute_pca(self):
        '''Compute k nearest neighbors from a matrix with fixed nodes

        Returns:
            list of lists with the first k or less indices of the neighbors for
            each free column. The length is N - n_fixed. For each now, there are
            less than k entries if no other column within the distance threshold
            were found, or if N < k.


        The algorithm proceeds as follows:
        0. take the log of the counts
        1. subtract the mean along the observation axis (N) and divide by the
        standard dev along the same axis
        2. calculate the weighted covariance matrix
        3. calculate normal PCA on that matrix
        '''
        matrix = self.matrix
        sizes = self.sizes
        n_atlas = self.n_atlas
        n_pcs = self.n_pcs

        # Test input arguments
        N, L = matrix.shape
        if len(sizes) != N:
            raise ValueError('Matrix and sizes dimensions do not match')
        if n_atlas >= N:
            raise ValueError('n_fixed larger or equal matrix number of columns')
        if n_pcs > min(L, N):
            raise ValueError('n_pcs greater than smaller matrix dimension, those eigenvalues are zero')

        # 0. take log
        matrix = np.log10(matrix + 0.1)

        # 1. standardize
        weights = 1.0 * sizes / sizes.sum()
        mean_w = weights @ matrix
        var_w = weights @ ((matrix - mean_w)**2)
        std_w = np.sqrt(var_w)
        Xnorm = (matrix - mean_w) / std_w

        # take care of non-varying components
        Xnorm[np.isnan(Xnorm)] = 0

        # 2. weighted covariance
        # This matrix has size L x L. Typically L ~ 500 << N, so the covariance
        # L x L is much smaller than N x N
        cov_w = np.cov(Xnorm.T, fweights=sizes)

        # 3. PCA
        # rvects columns are the right singular vectors
        evals, evects = np.linalg.eig(cov_w)
        # sort by decreasing eigenvalue (explained variance) and truncate
        ind = evals.argsort()[::-1][:n_pcs]
        # NOTE: we do not actually need the eigenvalues anymore
        lvects = np.real(evects.T[ind])

        # calculate right singular vectors given the left singular vectors
        # NOTE: this is true even if we truncated the PCA via n_pcs << L
        # rvects columns are the right singular vectors
        rvects = (lvects @ Xnorm.T).T

        # 4. expand embedded vectors to account for sizes
        # NOTE: this could be done by carefully tracking multiplicities
        # in the neighborhood calculation, but it's not worth it: the
        # amount of overhead memory used here is small because only a few
        # principal components are used
        Ne = int(np.sum(sizes))
        rvectse = np.empty((Ne, n_pcs), np.float32)
        cell_type_expanded = []
        i = 0
        n_fixed_expanded = 0
        for isi, size in enumerate(sizes):
            if isi < n_atlas:
                cte = self.cell_types_atlas[isi]
                n_fixed_expanded += int(size)
            else:
                cte = ''
            cell_type_expanded.extend([cte] * int(size))
            for j in range(int(size)):
                rvectse[i] = rvects[isi]
                i += 1
        cell_type_expanded = np.array(cell_type_expanded)

        self.pca_data = {
            'pcs': rvects,
            'pcs_expanded': rvectse,
            'cell_type': cell_type_expanded,
            'n_atlas': n_fixed_expanded,
            }

    def compute_similarity_graph(self):
        '''Compute similarity graph from the extended PC space

        1. calculate the distance matrix by expanding atlas columns
        2. calculate neighborhoods
        3. construct similarity graph from neighborhood lists
        '''
        from scipy.spatial.distance import cdist
        import igraph as ig

        sizes = self.sizes
        n_atlas = self.n_atlas
        k = self.n_neighbors
        kout = self.n_neighbors_out_of_atlas
        metric = self.distance_metric
        threshold = self.threshold_neighborhood
        rvects = self.pca_data['pcs']
        rvectse = self.pca_data['pcs_expanded']
        Ne = len(rvectse)

        # 5. calculate distance matrix and neighbors
        # we do it row by row, it costs a bit in terms of runtime but
        # has huge savings in terms of memory since we don't need the square
        # distance matrix
        n_fixede = int(np.sum(sizes[:n_atlas]))
        neighbors = []

        # Treat things within and outside of the atlas differently
        # Atlas neighbors
        i = 0
        for isi in range(n_atlas):
            # Find the nearest neighbors in the new data
            drow = cdist(rvects[[isi]], rvects[n_atlas:], metric=metric)[0]
            ind = np.argpartition(-drow, -kout)[-kout:]

            # Discard the ones beyond threshold
            ind = ind[drow[ind] <= threshold]

            # Indices are not sorted within ind, so we need to sort them
            # in descending order by distance (more efficient in the next step)
            ind = ind[np.argsort(drow[ind])]

            for ii in range(int(sizes[isi])):
                # Internal edges
                neis = list(range(i, i+int(sizes[isi])))
                # Remove self
                neis.remove(i+ii)
                # External edges
                neis.extend(list(ind + n_fixede))
                neighbors.append(neis)
            i += int(sizes[isi])

        # New data neighbors
        for i in range(n_fixede, Ne):
            drow = cdist(rvectse[[i]], rvectse, metric=metric)[0]

            # set distance to self as a high number, to avoid self
            drow[i] = drow.max() + 1

            # Find largest k negative distances (k neighbors)
            ind = np.argpartition(-drow, -k)[-k:]

            # Discard the ones beyond threshold
            ind = ind[drow[ind] <= threshold]

            # Indices are not sorted within ind, so we need to sort them
            # in descending order by distance (more efficient in the next step)
            ind = ind[np.argsort(drow[ind])]

            neighbors.append(list(ind))

        self.neighbors = neighbors

        # Construct graph from the lists of neighbors
        edges_d = set()
        for i, neis in enumerate(neighbors):
            for n in neis:
                edges_d.add(frozenset((i, n)))

        edges = [tuple(e) for e in edges_d]
        self.graph = ig.Graph(n=Ne, edges=edges, directed=False)

    def cluster_graph(self):
        '''Compute communities from a matrix with fixed nodes

        Returns:
            None, but Averages.membership is set as an array with
            size N - n_fixed with the atlas cell types of all cells from the
            new dataset.
        '''
        clu = ClusterWithAnnotations(
                self.graph,
                self.cell_types_atlas_extended,
                resolution_parameter=self.resolution_parameter,
                metric=self.clustering_metric,
            )
        self.membership = clu.fit_transform()

    def estimate_closest_atlas_cell_type(self):
        '''Estimate atlas cell type closest to each new cluster'''
        from scipy.spatial.distance import cdist

        # Use PC space
        rvectse = self.pca_data['pcs']
        n_atlas = self.pca_data['n_atlas']
        cell_types = self.pca_data['cell_type'][:n_atlas]
        L = rvectse.shape[1]

        # Extract atlas averages in PC space
        cell_types_atlas = np.unique(cell_types)
        rvectse_atlas = rvectse[:n_atlas]
        N = len(cell_types_atlas)
        avg_atl = np.empty((L, N), np.float32)
        for i, ct in enumerate(cell_types_atlas):
            # They are already replicates, take the first copy
            avg_atl[:, i] = rvectse_atlas[cell_types == ct][0]

        # Calculate averages for the new clusters
        cell_types_new = list(set(self.membership) - set(cell_types_atlas))
        rvectse_new = rvectse[n_atlas:]
        N = len(cell_types_new)
        avg_new = np.empty((L, N), np.float32)
        for i, ct in enumerate(cell_types_new):
            avg_new[:, i] = rvectse_new[self.membership == ct].mean(axis=0)

        # Calculate distance matrix between new and old in the high-dimensional
        # feature-selected space
        dmat = cdist(avg_new.T, avg_atl.T, metric='euclidean')

        # Pick the closest
        closest = np.argmin(dmat, axis=1)

        # Give it actual names
        closest = pd.Series(cell_types[closest], index=cell_types_new)

        return closest

    def embed(self, method='tsne', **kwargs):
        X = self.pca_data['pcs_expanded']
        index = list(self.cell_names_atlas_extended) + list(self.cell_names_newdata)

        if method == 'pca':
            emb = X[:, :2]
        elif method == 'tsne':
            from sklearn.manifold import TSNE
            kwargs['perplexity'] = kwargs.get('perplexity', 30)
            model = TSNE(
                n_components=2,
                **kwargs,
                )
            emb = model.fit_transform(X)
        elif method == 'umap':
            from umap import UMAP
            model = UMAP(
                n_components=2,
                **kwargs,
                )
            emb = model.fit_transform(X)

        res = pd.DataFrame(
            emb,
            index=index,
            columns=['Dimension 1', 'Dimension 2'],
            )
        res['CellType'] = list(self.cell_types_atlas_extended) + list(self.membership)
        res['Dataset'] = (['Atlas'] * self.n_atlas_extended) + (['New'] * self.n_newdata)

        return res



# vim: fdm=indent
# author:     Fabio Zanini
# date:       30/08/19
# content:    Test fetching atlases from
#             https://iosonofabio.github.io/atlas_averages/
import numpy as np
import pandas as pd
import anndata
from northstar import AtlasFetcher, subsample_atlas, average_atlas


def test_subsample_Darmanis_2015():
    af = AtlasFetcher()
    atlas = af.fetch_atlas('Darmanis_2015', kind='subsample')

    ss = subsample_atlas(
            atlas,
            n_cells=5,
            )

    assert(isinstance(ss, anndata.AnnData))
    assert(ss.shape[0] == 40)


def test_average_Darmanis_2015():
    af = AtlasFetcher()
    atlas = af.fetch_atlas('Darmanis_2015', kind='subsample')

    ss = average_atlas(
            atlas,
            )

    assert(isinstance(ss, anndata.AnnData))
    assert(ss.shape[0] == 8)



# vim: fdm=indent
# author:     Fabio Zanini
# date:       17/06/19
# content:    northstar entry point for import
from .averages import Averages
from .subsample import Subsample
from .fetch_atlas import AtlasFetcher
from .compress_atlas import subsample_atlas, average_atlas
from .cluster_with_annotations import ClusterWithAnnotations
from ._version import version

__all__ = [
    'Averages',
    'Subsample',
    'AtlasFetcher',
    'ClusterWithAnnotations',
    'subsample_atlas',
    'average_atlas',
    'version',
    ]



# vim: fdm=indent
# author:     Fabio Zanini
# date:       5/08/19
# content:    Cluster with annotations. This is used in Subsample and can be
#             used independently with external similarity graphs and annotations.
__all__ = ['ClusterWithAnnotations']

import numpy as np
import scipy as sp
import igraph as ig
import leidenalg


class ClusterWithAnnotations(object):
    '''Cluster a similarity graph of cells with fixed atlas annotations'''

    def __init__(
            self,
            graph,
            annotations,
            resolution_parameter=0.001,
            metric='cpm',
            ):
        '''Prepare the clustering with annotations


        Args:
            graph (igraph.Graph or adjecency matrix): the
                similarity graph connecting cells. The first N cells are
                considered part of the atlas, where N is the length of the
                annotations argument.

            annotations (list or numpy ndarray): cell type annotations for
                the atlas cells, assumed to be the first one in the graph.

            resolution_parameter (float): number between 0 and 1 that sets
             how easy it is for the clustering algorithm to make new clusters

            metric (str): 'cpm' (default, Cell Potts Model) or 'modularity'.
                Sets the type of partition used in the clustering step.

        '''
        self.graph = graph
        self.annotations = annotations
        self.resolution_parameter = resolution_parameter
        self.metric = metric

    def _parse_graph(self):
        '''Convert whatever input for the graph attribute into igraph.Graph'''
        arg = self.graph
        if isinstance(arg, ig.Graph):
            return

        if isinstance(arg, np.ndarray) and (arg.shape[0] == arg.shape[1]):
            n_nodes = arg.shape[0]
            edges = list(zip(*(arg.nonzero())))
        elif sp.sparse.issparse(arg):
            arg = arg.tocoo()
            n_nodes = arg.shape[0]
            edges = list(zip(arg.row, arg.col))
        else:
            raise ValueError('Graph format not recognized')

        self.graph = ig.Graph(n=n_nodes, edges=edges)

    def fit(self):
        '''Compute communities from a matrix with fixed nodes

        Returns:
            None, but the membership attribute is set as an array of int with
            size N - n_fixed with the community/cluster membership of all
            columns except the first n fixed ones.
        '''
        self._parse_graph()

        aa = self.annotations
        n_fixed = len(aa)
        g = self.graph
        N = g.vcount()

        opt = leidenalg.Optimiser()
        is_membership_fixed = [int(i < n_fixed) for i in range(N)]

        # NOTE: initial membership is singletons except for atlas nodes, which
        # get the membership they have.
        aau = list(np.unique(aa))
        aaun = len(aau)
        initial_membership = []
        for j in range(N):
            if j < n_fixed:
                mb = aau.index(aa[j])
            else:
                mb = aaun + (j - n_fixed)
            initial_membership.append(mb)

        if self.metric == 'cpm':
            partition = leidenalg.CPMVertexPartition(
                    g,
                    resolution_parameter=self.resolution_parameter,
                    initial_membership=initial_membership,
                    )
        elif self.metric == 'modularity':
            partition = leidenalg.ModularityVertexPartition(
                    g,
                    resolution_parameter=self.resolution_parameter,
                    initial_membership=initial_membership,
                    )
        else:
            raise ValueError(
                'clustering_metric not understood: {:}'.format(self.metric))

        # Run modified Leiden here
        opt.optimise_partition(partition,
                               is_membership_fixed=is_membership_fixed)

        # Exctract result
        membership = partition.membership[n_fixed:]

        # Convert the known cell types
        lstring = len(max(aau, key=len))
        self.membership = np.array(
                [str(x) for x in membership],
                dtype='U{:}'.format(lstring))
        for i, ct in enumerate(aau):
            self.membership[self.membership == str(i)] = ct

    def fit_transform(self):
        '''Cluster graph and return memberships'''
        self.fit()
        return self.membership



# vim: fdm=indent
# author:     Fabio Zanini
# date:       5/08/19
# content:    Atlas subsampling
__all__ = ['Subsample']

import os
import warnings
import numpy as np
import pandas as pd
import scipy as sp
from anndata import AnnData, concat
import leidenalg


# scanpy can be skipped if requested
if os.getenv('NORTHSTAR_SKIP_SCANPY') is not None:
    scanpy = None
else:
    try:
        import scanpy
    except ImportError:
        scanpy = None

from .fetch_atlas import AtlasFetcher
from .cluster_with_annotations import ClusterWithAnnotations


class Subsample(object):
    '''Annotate new cell types using an even subsample of an atlas'''

    def __init__(
            self,
            atlas,
            n_features_per_cell_type=30,
            n_features_overdispersed=500,
            features_additional=None,
            n_pcs=20,
            n_neighbors=10,
            n_neighbors_external=0,
            external_neighbors_mutual=False,
            distance_metric='correlation',
            threshold_neighborhood=0.8,
            threshold_neighborhood_external=0.8,
            clustering_metric='cpm',
            resolution_parameter=0.003,
            normalize_counts=True,
            join='keep_first',
            atlas_annotation_column='CellType',
            fraction_similar_types=0.1,
            n_features_similar_types=0,
            ):
        '''Prepare the model for cell annotation

        Args:
            atlas (str, list of str, list of dict, dict, or AnnData): cell
             atlas to use. Generally there are two kind of choices:

             The first possibility selects the corresponding cell atlas or
             atlases from northstar's online list. The names of currently
             available dataset is here:

             https://github.com/iosonofabio/atlas_averages/blob/master/table.tsv

             (check the first column for atlas names). If a list of
             str, multiple atlases will be fetched and combined. Only features
             that are in all atlases will be kept. If you use this feature, be
             careful to not mix atlases from different species. If a list of
             dict, it merges atlases as above but you can specify what cell
             types to fetch from each atlas. Each element of the list must be a
             dict with two key-value pairs: 'atlas_name' is the atlas name, and
             'cell_types' must be a list of cell types to retain. Example:
             atlas=[{'atlas_name': 'Enge_2017', 'cell_types': ['alpha']}] would
             load the atlas Enge_2017 and only retain alpha cells. You can also
             use a dict to specify a single atlas and to retain only certain cell
             types. The format is as above, e.g. to select only alpha cells
             from Enge_2017 you can use:

             atlas={'atlas_name': 'Enge_2017', 'cell_types': ['alpha']}.

             The second possibility is to use a custom atlas (e.g. some
             unpublished data). 'atlas' must be an AnnData object with cells as
             rows and genes as columns and one cell metadata column
             describing the cell type. If the "atlas_annotation_column" is not
             set,

                 adata.obs['CellType']

             must exist. If you set that option manually, the obs colunm with
             that name must exist instead.

            n_features_per_cell_type (int): number of features marking each
             fixed column (atlas cell type).

            n_features_overdispersed (int): number of unbiased, overdispersed
             features from the last N - n_fixed columns.

            features_additional (list of str or None): additional features to
             keep on top of automatic selection. The argument 'features' takes
             priority over this one.

            n_pcs (int): number of principal components to keep in the PCA

            n_neighbors (int): number of neighbors in the similarity graph

            n_neighbors_external (int): number of additional neighbors in the
                similarity graph that must be in the training data. This option
                is provided to force a degree of connectivity between the
                new data and the atlas (default: 0).

            external_neighbors_mutual (bool): when using external neighbors,
                only add them if they are mutual neighbors. This option is used
                to balance out a little the effect of n_neighbors_external > 0.

            distance_metric (str): metric to use as distance. It should be a
             metric accepted by scipy.spatial.distance.cdist.

            threshold_neighborhood (float): do not consider distances larger
             than this as neighbors

            threshold_neighborhood_external (float): do not consider distances
                larger than this for external neighbors (if requested)

            clustering_metric (str): 'cpm' (default, Cell Potts Model) or
             'modularity'. Sets the type of partition used in the clustering
             step.

            resolution_parameter (float): number between 0 and 1 that sets
             how easy it is for the clustering algorithm to make new clusters

            normalize_counts (bool): whether to renormalize the counts at the
             merging stage to make sure atlas and new data follow the same
             normalization. Be careful if you turn this off.

            join (str): must be 'keep_first', 'union', or 'intersection'. This
             argument is used when sourcing multiple atlases and decides what
             to do with features that are not present in all atlases.
             'keep_first' keeps the features in the first atlas and pads the
             other atlases with zeros, 'union' pads every atlas that is missing
             a feature and 'intersection' only keep features that are in all
             atlases.

            atlas_annotation_column (str): The name of the obs column with
             the atlas annotations (default: 'CellType').

            fraction_similar_types (float): The fraction of atlas cell types
             that are similar enough to require additional pairwise markers.

            n_features_similar_types (int): The number of features used for
             each pairwise comparison between similar atlas cell types.
        '''

        self.atlas = atlas
        self.n_features_per_cell_type = n_features_per_cell_type
        self.n_features_overdispersed = n_features_overdispersed
        self.features_additional = features_additional
        self.n_pcs = n_pcs
        self.n_neighbors = n_neighbors
        self.n_neighbors_external = n_neighbors_external
        self.external_neighbors_mutual = external_neighbors_mutual
        self.distance_metric = distance_metric
        self.threshold_neighborhood = threshold_neighborhood
        self.threshold_neighborhood_external = threshold_neighborhood_external
        self.clustering_metric = clustering_metric
        self.resolution_parameter = resolution_parameter
        self.normalize_counts = normalize_counts
        self.join = join
        self.atlas_annotation_column = atlas_annotation_column
        self.fraction_similar_types = fraction_similar_types
        self.n_features_similar_types = n_features_similar_types

    def fit(self, new_data):
        '''Run with subsamples of the atlas

        Args:
            new_data (pandas.DataFrame or anndata.AnnData): the new data to be
             clustered. If a dataframe, t must have features as rows and
             cell names as columns (as in loom files). anndata uses the opposite
             convention, so it must have cell names as rows (obs_names) and
             features as columns (var_names) and this class will transpose it.

        Returns:
            None, but this instance of Subsample acquired the property
            `membership` containing the cluster memberships (cell types) of the
            columns except the first n_fixed. The first n_fixed columns are
            assumes to have distinct memberships in the range [0, n_fixed - 1].
        '''
        self.new_data = new_data

        self._check_init_arguments()
        self.fetch_atlas_if_needed()
        self.compute_feature_intersection()
        self._check_feature_intersection()
        self.prepare_feature_selection()
        self.select_features()
        self._check_feature_selection()
        self.merge_atlas_newdata()
        self.compute_pca()
        self.compute_similarity_graph()
        self.cluster_graph()

    def fit_transform(self, new_data):
        '''Run with subsamples of the atlas and return the cell types

        Args:
            new_data (pandas.DataFrame or anndata.AnnData): the new data to be
             clustered. If a dataframe, t must have features as rows and
             cell names as columns (as in loom files). anndata uses the opposite
             convention, so it must have cell names as rows (obs_names) and
             features as columns (var_names) and this class will transpose it.

        Returns:
            the cluster memberships (cell types) of the
            columns except the first n_fixed. The first n_fixed columns are
            assumes to have distinct memberships in the range [0, n_fixed - 1].
        '''
        self.fit(new_data)
        return self.membership

    def _check_init_arguments(self):
        # Check atlas
        at = self.atlas

        if isinstance(at, str):
            pass
        elif isinstance(at, list) or isinstance(at, tuple):
            for elem in at:
                if isinstance(elem, str):
                    pass
                elif isinstance(elem, dict):
                    if 'atlas_name' not in elem:
                        raise ValueError('List of atlases: format incorrect')
                    if 'cell_types' not in elem:
                        raise ValueError('List of atlases: format incorrect')
                else:
                    raise ValueError('List of atlases: format incorrect')
        elif isinstance(at, dict) and ('atlas_name' in at) and \
                ('cell_types' in at):
            pass

        elif isinstance(at, AnnData):
            if self.atlas_annotation_column not in at.obs:
                raise AttributeError('atlas does not have a "{:}" obs column'.format(
                    self.atlas_annotation_column))

        else:
            raise ValueError('Atlas not formatted correctly')

        # Convert new data to anndata if needed
        nd = self.new_data
        if isinstance(nd, AnnData):
            pass
        elif isinstance(nd, pd.DataFrame):
            # AnnData uses features as columns, so transpose and convert
            # (the assumption is that in the DataFrame convention, rows are
            # features)
            nd = AnnData(
                X=nd.values.T,
                obs={'CellID': nd.columns.values},
                var={'Gene Name': nd.index.values},
                )
            self.new_data = nd

        # New data could be too small to do PCA
        n_newcells, n_newgenes = self.new_data.shape
        if n_newgenes < self.n_pcs:
            warnings.warn(
                ('The number of features in the new data is lenn than ' +
                 'the number of PCs, so northstar might give inaccurate ' +
                 'results'))

        if n_newcells < self.n_pcs:
            warnings.warn(
                ('The number of cells in the new data is lenn than ' +
                 'the number of PCs, so northstar might give inaccurate ' +
                 'results'))

        if min(n_newgenes, n_newcells) < self.n_pcs:
            warnings.warn('Reducing the number of PCs to {:}'.format(
                min(n_newgenes, n_newcells)))
            self.n_pcs = min(n_newgenes, n_newcells)

        # New data could be too small for knn
        if n_newcells < self.n_neighbors + 1:
            warnings.warn(
                ('The number of cells in the new data is less than the ' +
                 'number of neighbors requested for the knn: reducing the ' +
                 'number of graph neighbors to {:}'.format(
                     max(1, n_newcells - 1)),
                 ))
            self.n_neighbors = max(1, n_newcells - 1)

        nf1 = self.n_features_per_cell_type
        nf2 = self.n_features_overdispersed
        nf3 = self.features_additional
        nf4 = self.n_features_similar_types
        if (not isinstance(nf1, int)) or (nf1 < 0):
            raise ValueError('n_features_per_cell_type must be an int >= 0')
        if (not isinstance(nf2, int)) or (nf2 < 0):
            raise ValueError('n_features_overdispersed must be an int >= 0')
        if (not isinstance(nf4, int)) or (nf4 < 0):
            raise ValueError('n_features_similar_types must be an int >= 0')

        if (nf1 < 1) and (nf2 < 1) and (nf3 < 1):
            raise ValueError('No features selected')

    def _check_feature_intersection(self):
        L = len(self.features_ovl)
        if L == 0:
            raise ValueError(
                ('No overlapping features in atlas and new data, are gene ' +
                 'names correct for this species?'))
        if L < 50:
            warnings.warn(
                ('Only {:} overlapping features found in atlas and new ' +
                 'data'.format(L)))

    def _check_feature_selection(self):
        L = len(self.features)
        if L == 0:
            raise ValueError(
                ('No features survived selection, check nortstar parameters'))
        if L < self.n_pcs:
            warnings.warn(
                ('Only {0} features selected, reducing PCA to {0} ' +
                 'components'.format(L)))
            self.n_pcs = L

    def fetch_atlas_if_needed(self):
        '''Fetch atlas(es) if needed'''
        at = self.atlas

        if isinstance(at, str):
            self.atlas = AtlasFetcher().fetch_atlas(
                    at,
                    kind='subsample',
                    )

        elif isinstance(at, list) or isinstance(self.atlas, tuple):
            self.atlas = AtlasFetcher().fetch_multiple_atlases(
                    at,
                    kind='subsample',
                    join=self.join,
                    )

        elif isinstance(at, dict) and ('atlas_name' in at) and \
                ('cell_types' in at):
            self.atlas = AtlasFetcher().fetch_atlas(
                    at['atlas_name'],
                    kind='subsample',
                    cell_types=at['cell_types'],
                    )

    def compute_feature_intersection(self):
        '''Calculate the intersection of features between atlas and new data'''
        # Intersect features
        self.features_atlas = self.atlas.var_names.values
        self.features_newdata = self.new_data.var_names.values
        self.features_ovl = np.intersect1d(
                self.features_atlas,
                self.features_newdata,
                )

    def prepare_feature_selection(self):
        # Cell names and types
        self.cell_types_atlas = self.atlas.obs[self.atlas_annotation_column].values
        self.cell_names_atlas = self.atlas.obs_names.values
        self.cell_names_newdata = self.new_data.obs_names.values

        # Numbers
        self.n_atlas = self.atlas.shape[0]
        self.n_newdata = self.new_data.shape[0]
        self.n_total = self.n_atlas + self.n_newdata

    def select_features(self):
        '''Select features that define heterogeneity of the atlas and new data

        Returns:
            ndarray of feature names.
        '''
        from scipy.spatial.distance import pdist, squareform

        features_atlas = self.features_atlas
        features_newdata = self.features_newdata
        features_ovl = list(self.features_ovl)
        features_add = self.features_additional

        nf1 = self.n_features_per_cell_type
        nf2 = self.n_features_overdispersed

        cell_types = self.cell_types_atlas
        cell_typesu = list(np.unique(cell_types))

        # This is the output data structure
        features = set()

        # Atlas markers, one vs rest
        self.features_atlas_selected = {}
        if len(cell_typesu) > 1:
            avg = np.zeros((len(cell_typesu), len(features_atlas)), np.float32)
            matrix = self.atlas.X
            for iau, au in enumerate(cell_typesu):
                ic1 = (cell_types == au).nonzero()[0]
                ge1 = np.asarray(matrix[ic1].mean(axis=0))
                if sp.sparse.issparse(matrix):
                    ge1 = ge1[0]
                avg[iau] = ge1
            self.atlas_averages = pd.DataFrame(avg, index=cell_typesu, columns=features_atlas)[features_ovl]

            for iau, au in enumerate(cell_typesu):
                ge1 = avg[iau]
                ic2 = (cell_types != au).nonzero()[0]
                ge2 = np.asarray(matrix[ic2].mean(axis=0))
                if sp.sparse.issparse(matrix):
                    ge2 = ge2[0]
                fold_change = np.log2(ge1 + 0.1) - np.log2(ge2 + 0.1)
                tmp = np.argsort(fold_change)[::-1]
                ind_markers_atlas = []
                for i in tmp:
                    if features_atlas[i] in features_ovl:
                        ind_markers_atlas.append(i)
                    if len(ind_markers_atlas) == nf1:
                        break
                # Add atlas markers
                fa_au = set(features_atlas[ind_markers_atlas])
                features |= fa_au
                self.features_atlas_selected[au] = fa_au

        # Atlas markers, pairwise for the closest categories
        nmark = self.n_features_similar_types
        nct = int(len(cell_typesu) * self.fraction_similar_types)
        if (nmark > 0) and (nct > 0):
            self.similar_types = []
            dmat = squareform(pdist(avg))
            # Sort the pairs of cell types by distance, skip the zeros, and
            # remember the matrix is symmetric
            tmp = np.argsort(dmat.ravel())[dmat.shape[0]:][:2*nct:2]
            idx1, idx2 = np.unravel_index(tmp, dmat.shape)
            for iau1, iau2 in zip(idx1, idx2):
                self.similar_types.append((cell_typesu[iau1], cell_typesu[iau2]))
                ge1 = avg[iau1]
                ge2 = avg[iau2]
                # NOTE: in this case, we do not care about up- or down-regulated
                # The purpose is simply to distinguish the two cell types
                # FIXME: actually we do, because of dropout
                fold_change = np.abs(np.log2(ge1 + 0.1) - np.log2(ge2 + 0.1))
                tmp = np.argsort(fold_change)[::-1]
                # Add top markers in the intersection
                ind_markers_atlas = []
                for i in tmp:
                    if features_atlas[i] in features_ovl:
                        ind_markers_atlas.append(i)
                    if len(ind_markers_atlas) == nmark:
                        break
                fa_au = set(features_atlas[ind_markers_atlas])
                features |= fa_au

        # Unbiased on new data
        if nf2 > 0:
            if nf2 >= len(features_ovl):
                features |= set(features_ovl)
            else:
                matrix = self.new_data.X
                nd_mean = np.asarray(matrix.mean(axis=0))[0]
                if sp.sparse.issparse(matrix):
                    m2 = matrix.copy()
                    m2.data **= 2
                else:
                    m2 = matrix ** 2
                nd_var = np.asarray(m2.mean(axis=0))[0] - nd_mean**2
                fano = (nd_var + 1e-10) / (nd_mean + 1e-10)
                tmp = np.argsort(fano)[::-1]
                ind_ovd_newdata = []
                for i in tmp:
                    if features_newdata[i] in features_ovl:
                        ind_ovd_newdata.append(i)
                    if len(ind_ovd_newdata) == nf2:
                        break
                # Add overdispersed features
                features |= set(features_newdata[ind_ovd_newdata])

        # Additional features
        if features_add is not None:
            features |= (set(features_add) & set(features_ovl))

        self.features = np.array(list(features))

    def merge_atlas_newdata(self):
        '''Merge atlas data and the new data after feature selection

        NOTE: is self.normalize is True, the merged count matrix is normalized
        by 1 million total counts.
        '''

        features = self.features
        L = len(features)
        N = self.n_total
        N1 = self.n_atlas

        # Find the feature indices for atlas
        ind_features_atlas = pd.Series(
            np.arange(len(self.features_atlas)),
            index=self.features_atlas,
            ).loc[features].values

        # Find the feature indices for new data
        ind_features_newdata = pd.Series(
            np.arange(len(self.features_newdata)),
            index=self.features_newdata,
            ).loc[features].values

        # This is the largest memory footprint of northstar
        # Try using sparse matrices if the data is using them already
        adata_merge = concat(
            [self.atlas[:, ind_features_atlas],
             self.new_data[:, ind_features_newdata]],
        )
        self.adata_merge = adata_merge

        if self.normalize_counts:
            if scanpy is not None:
                scanpy.pp.normalize_total(
                    adata_merge,
                    target_sum=1e6,
                    inplace=True,
                )
            else:
                # The normalization function should not set pseudocounts
                matrix = adata_merge.X
                coverage = matrix.sum(axis=1)
                # Note if cells have no reads
                cells_no_reads = coverage == 0
                if cells_no_reads.sum():
                    print('Some cells have no reads in the feature selected space')
                coverage[cells_no_reads] = 1
                matrix = 1e6 * (matrix.T / coverage).T
                self.adata_merge.X = matrix

    def compute_pca(self):
        '''Compute k nearest neighbors from a matrix with fixed nodes

        Returns:
            list of lists with the first k or less indices of the neighbors for
            each free column. The length is N - n_fixed. For each now, there are
            less than k entries if no other column within the distance threshold
            were found, or if N < k.


        The algorithm proceeds as follows:
        1. whiten the matrix, i.e. subtract the mean along
        the observation axis (N) and divide by the standard dev along the same axis
        2. calculate the weighted covariance matrix
        3. calculate normal PCA on that matrix
        '''
        from sklearn.decomposition import PCA

        matrix = self.adata_merge.X
        n_fixed = self.n_atlas
        n_pcs = self.n_pcs

        # Test input arguments
        N, L = matrix.shape
        if n_fixed >= N:
            raise ValueError(
                'n_fixed larger or equal matrix number of columns')
        if n_pcs > min(L, N):
            raise ValueError(
                ('n_pcs greater than smaller matrix dimension, the ' +
                 'remaining eigenvalues will be zero'))

        # 0. take log
        if scanpy is not None:
            matrix = scanpy.pp.log1p(matrix, base=10)
        else:
            matrix = np.log1p(matrix) / np.log(10)

        # TODO: make dense here, peek at scanpy to see how they do it
        if isinstance(matrix, sp.sparse.spmatrix):
            matrix = matrix.toarray()

        # 1. whiten
        Xnorm = (matrix - matrix.mean(axis=0)) / matrix.std(axis=0, ddof=0)

        # take care of non-varying components
        Xnorm[np.isnan(Xnorm)] = 0

        # 2. PCA
        pca = PCA(n_components=n_pcs)
        # rvects columns are the right singular vectors
        rvects = pca.fit_transform(Xnorm)

        self.pca_data = {
            'pcs': rvects,
            'cell_type': np.array(list(self.cell_types_atlas) + [''] * (N - n_fixed)),
            'n_atlas': n_fixed,
            }

    def compute_similarity_graph(self):
        '''Compute similarity graph from the extended PC space

        1. calculate the distance matrix
        2. calculate neighborhoods
        3. construct similarity graph from neighborhood lists
        '''
        from scipy.spatial.distance import pdist, cdist, squareform
        import igraph as ig

        matrix = self.adata_merge.X
        k = self.n_neighbors
        ke = self.n_neighbors_external
        metric = self.distance_metric
        threshold = self.threshold_neighborhood
        threshold_external = self.threshold_neighborhood_external
        rvects = self.pca_data['pcs']
        N, L = matrix.shape
        N1 = self.n_atlas

        # If scanpy is available, use it for the bulk connections.
        # External links to atlas canbe done by hand because they are fewer
        if scanpy is not None:
            self.adata_merge.obsm['X_pca'] = rvects
            scanpy.pp.neighbors(
                    self.adata_merge,
                    n_neighbors=k,
                    n_pcs=rvects.shape[1],
                    use_rep='X_pca',
                    metric=metric,
                )
            # NOTE: we'll alter the sparse structure, so use lil
            # That reflects in the raw data operations below
            dmat = self.adata_merge.obsp['distances'].tolil()
            cmat = self.adata_merge.obsp['connectivities'].tolil()

            # Additional neighbors from the atlas if requested
            if ke > 0:
                dmat2 = cdist(rvects[N1:], rvects[:N1], metric=metric)
                # NOTE: Offset i by N1, dmat is only the rectangular slice
                for i, drow in enumerate(dmat2, N1):
                    ind2 = np.argpartition(-drow, -ke)[-ke:]
                    ind2 = ind2[drow[ind2] <= threshold_external]
                    ind2 = ind2[np.argsort(drow[ind2])]
                    ind2 = list(ind2)

                    # Require mutual neighbors if requested
                    if self.external_neighbors_mutual:
                        ind2 = [j for j in ind2 if dmat[j, i] > 0]

                    # Skip existing edges
                    ind2 = [j for j in ind2 if dmat[i, j] == 0]
                    for j in ind2:
                        dmat[i, j] = drow[j]

                # Fix connectivity
                # FIXME: double check connectivity algo (e.g. exp a la UMAP)
                for i, drow in enumerate(dmat[N1:], N1):
                    dat = np.asarray(drow.data[0])
                    # A la UMAP, subtract shortest distance (local warping)...
                    dat -= dat.min()
                    # ...and then exponentiate
                    eta = 1.  # FIXME: find decent number
                    dat = np.exp(-dat / eta)

                    # NOTE: no need to offset j, since the atlas comes first
                    for j, conn in zip(drow.rows[0], dat):
                        cmat[i, j] = conn

            self.adata_merge.obsp['distances'] = dmat.tocsr()
            self.adata_merge.obsp['connectivities'] = cmat.tocsr()

            # Build graph object
            cmat = cmat.tocoo()
            edges = [(i, j) for i, j in zip(cmat.row, cmat.col)]
            weights = cmat.data
            # NOTE: perhaps we should start using weights...
            self.graph = ig.Graph(
                n=N, edges=edges, directed=False,
                edge_attrs={'weight': weights},
            )

        else:
            # 1. calculate distance matrix
            # rvects is N x n_pcs. Let us calculate the end that includes only the free
            # observations and then call cdist which kills the columns. The resulting
            # matrix has dimensions N x N
            if N <= 10000:
                dmat = squareform(pdist(rvects, metric=metric))
                dmax = dmat.max()
            # For large data sets, compute distances row by row to save memory
            else:
                dmat = None
                dmax = None

            # TODO: we could use some accurate stochastic algo here, such as LSH
            # or the one scanpy is using. It makes no big difference whether the
            # graph is exact or not

            # 2. calculate neighbors
            neighbors = [[] for n1 in range(N)]
            for i, nbi in enumerate(neighbors):
                if dmat is not None:
                    drow = dmat[i]
                    # set distance to self as a high number, to avoid self
                    drow[i] = dmax + 1
                else:
                    drow = cdist(rvects, rvects[[i]], metric=metric)[:, 0]
                    dmax = drow.max()

                # Find largest k negative distances (k neighbors)
                ind = np.argpartition(-drow, -k)[-k:]

                # Discard the ones beyond threshold
                ind = ind[drow[ind] <= threshold]

                # Indices are not sorted within ind, we sort them to help
                # debugging even though it is not needed
                ind = ind[np.argsort(drow[ind])]
                ind = list(ind)

                # Additional neighbors from the atlas if requested
                if (ke > 0) and (i >= N1):
                    drow = drow[:N1]
                    ind2 = np.argpartition(-drow, -ke)[-ke:]
                    ind2 = ind2[drow[ind2] <= threshold_external]
                    ind2 = ind2[np.argsort(drow[ind2])]
                    ind2 = list(ind2)

                    # Require mutual neighbors if requested
                    if self.external_neighbors_mutual:
                        ind2 = [x for x in ind2 if i in neighbors[x]]

                    for i2 in ind2:
                        if i2 not in ind:
                            ind.append(i2)

                nbi.extend(ind)

            self.neighbors = neighbors

            # Construct graph from the lists of neighbors
            edges_d = set()
            for i, neis in enumerate(neighbors):
                for n in neis:
                    edges_d.add(frozenset((i, n)))

            edges = [tuple(e) for e in edges_d]
            self.graph = ig.Graph(n=N, edges=edges, directed=False)

    def cluster_graph(self):
        '''Compute communities from a matrix with fixed nodes

        Returns:
            None, but Subsample.membership is set as an array of int with
            size N - n_fixed with the community/cluster membership of all
            columns except the first n_fixed ones.
        '''
        clu = ClusterWithAnnotations(
                self.graph,
                self.cell_types_atlas,
                resolution_parameter=self.resolution_parameter,
                metric=self.clustering_metric,
            )
        self.membership = clu.fit_transform()

        # Annotate the AnnData object with source and membership
        N1 = self.n_atlas
        self.adata_merge.obs['northstar_source'] = 'new'
        self.adata_merge.obs['northstar_source'].iloc[:N1] = 'atlas'
        source = self.adata_merge.obs['northstar_source']
        source = pd.Series(data=pd.Categorical(source), index=source.index)
        self.adata_merge.obs['northstar_source'] = source

        annos = self.adata_merge.obs[self.atlas_annotation_column]
        annos.iloc[N1:] = self.membership
        annos = pd.Series(data=pd.Categorical(annos), index=annos.index)
        self.adata_merge.obs[self.atlas_annotation_column] = annos

    def estimate_closest_atlas_cell_type(self):
        '''Estimate atlas cell type closest to each new cluster'''
        from scipy.spatial.distance import cdist

        matrix = self.adata_merge.X
        n_fixed = self.n_atlas
        metric = self.distance_metric
        cell_types = self.cell_types_atlas

        # Calculate averages for the new clusters
        ct_new = list(set(self.membership) - set(cell_types))
        N = len(ct_new)
        L = matrix.shape[1]
        avg_new = np.empty((N, L), np.float32)
        for i, ct in enumerate(ct_new):
            avg_new[:, i] = matrix[self.membership == ct].mean(axis=0)

        avg_atl = np.empty((len(cell_types), L), np.float32)
        for i, ct in enumerate(cell_types):
            avg_atl[:, i] = matrix[self.membership[:n_fixed] == ct].mean(axis=0)

        # Calculate distance matrix between new and old in the high-dimensional
        # feature-selected space
        dmat = cdist(avg_new.T, avg_atl.T, metric=metric)

        # Pick the closest
        closest = np.argmin(dmat, axis=1)

        # Give it actual names
        closest = pd.Series(cell_types[closest], index=ct_new)

        return closest

    def embed(self, method='tsne', n_components=2, **kwargs):
        '''Embed atlas and new data into a low-dimensional space


        Args:
            method (str): One of 'tsne', 'pca', and 'umap'.
            n_components (int): number of dimensions of the embedding
            **kwargs: for 'tsne' and 'umap', keyword argument to their
                constructors

        Returns:
            pd.DataFrame with the low-dimensional coordinates
        '''
        X = self.pca_data['pcs']
        index = list(self.cell_names_atlas) + list(self.cell_names_newdata)
        columns = ['Dimension {:}'.format(i+1) for i in range(n_components)]

        if method == 'pca':
            emb = X[:, :n_components]
        elif method == 'tsne':
            from sklearn.manifold import TSNE
            kwargs['perplexity'] = kwargs.get('perplexity', 30)
            model = TSNE(
                n_components=n_components,
                **kwargs,
                )
            emb = model.fit_transform(X)
        elif method == 'umap':
            from umap import UMAP
            model = UMAP(
                n_components=n_components,
                **kwargs,
                )
            emb = model.fit_transform(X)

        res = pd.DataFrame(
            emb,
            index=index,
            columns=columns,
            )
        res[self.atlas_annotation_column] = list(self.cell_types_atlas) + list(self.membership)
        res['Dataset'] = (['Atlas'] * self.n_atlas) + (['New'] * self.n_newdata)

        return res



# vim: fdm=indent
# author:     Fabio Zanini
# date:       17/06/19
# content:    Test the algorithm on same artificial data
import numpy as np
import pandas as pd
from northstar import Subsample, AtlasFetcher


def test_run_within_atlas():
    aname = 'Darmanis_2015'
    atlas = AtlasFetcher().fetch_atlas(
            aname, kind='subsample')
    matrix = atlas.copy()
    cell_types = atlas.obs['CellType'].values

    sa = Subsample(aname)
    sa.fit(matrix)

    # Small samples are a bit tricky, one cell can tip the balance
    assert((cell_types == sa.membership).mean() >= 0.85)


def test_run_across_atlas():
    atlas = AtlasFetcher().fetch_atlas(
            'Enge_2017', kind='subsample')
    matrix = atlas.copy()
    cell_types = atlas.obs['CellType'].values

    sa = Subsample(
        'Baron_2016',
        n_pcs=25,
        n_features_per_cell_type=3,
        n_features_overdispersed=200,
        )
    sa.fit(matrix)

    # Nobody's perfect
    # Baron annotates Stellate cells more accurately, so we skip them
    assert((cell_types == sa.membership)[:60].mean() >= 0.4)


if __name__ == '__main__':

    test_run_across_atlas()



import os
from setuptools import setup, find_packages



def read(fname):
    this_directory = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(this_directory, fname)) as f:
        out = f.read()
    return out


def update_version():
    ver = read('northstar/_version.py').rstrip('\n').split()[-1].strip('"')
    return ver


def get_readme():
    return read('README.md')


setup(
    name="northstar",
    version=update_version(),
    author="Fabio Zanini",
    author_email="fabio.zanini@fastmail.fm",
    description="Single cell type annotation guided by cell atlases, with freedom to be queer.",
    license="MIT",
    keywords="graph semi-supervised",
    url="https://github.com/northstaratlas/northstar",
    packages=['northstar'] + ['northstar.' + s for s in find_packages(where='northstar')],
    long_description=get_readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=[
        'numpy',
        'pandas',
        'scikit-learn',
        'python-igraph>=0.8.0',
        'leidenalg>=0.8.0',
        'anndata',
    ],
    setup_requires=[
        'numpy',
        'pandas',
        'scikit-learn',
        'python-igraph>=0.8.0',
        'leidenalg>=0.8.0',
        'anndata',
    ],
    extras_require={
        'atlas-fetcher': [
            'requests',
            'loompy',
        ]
    },
)



# vim: fdm=indent
# author:     Fabio Zanini
# date:       17/06/19
# content:    Test the algorithm on same artificial data
import numpy as np
import pandas as pd
from northstar import Averages, AtlasFetcher


def test_constructor():
    sa = Averages(
            'Enge_2017',
            n_neighbors=2,
            threshold_neighborhood=0.8,
            n_pcs=2,
            distance_metric='correlation',
            n_neighbors_out_of_atlas=1
            )

    assert(sa is not None)


def test_arguments():
    matrix = pd.DataFrame(
        index=['INS', 'GCG', 'PPY'],
        columns=['cell1', 'cell2', 'cell3'],
        data=[
            [2302, 123, 0],
            [0, 5034, 6453],
            [0, 0, 1]],
        )

    sa = Averages(
            'Enge_2017',
            n_neighbors=2,
            threshold_neighborhood=0.8,
            n_pcs=2,
            distance_metric='correlation',
            n_neighbors_out_of_atlas=1
            )

    sa.new_data = matrix
    sa._check_init_arguments()
    assert(sa is not None)


def test_merge_small():
    matrix = pd.DataFrame(
        index=['INS', 'GCG', 'PPY'],
        columns=['cell1', 'cell2', 'cell3'],
        data=[
            [2302, 123, 0],
            [0, 5034, 6453],
            [0, 0, 1]],
        )
    n_pcs = 2
    k = 1
    threshold = 0.8

    sa = Averages(
            'Enge_2017',
            n_neighbors=k,
            threshold_neighborhood=threshold,
            n_pcs=n_pcs,
            distance_metric='correlation',
            n_neighbors_out_of_atlas=1
            )

    sa.new_data = matrix
    sa._check_init_arguments()
    sa.fetch_atlas_if_needed()
    sa.compute_feature_intersection()
    sa._check_feature_intersection()
    sa.prepare_feature_selection()
    sa.select_features()
    sa._check_feature_selection()
    sa.merge_atlas_newdata()

    assert(sa is not None)


def test_neighbors_small():
    matrix = pd.DataFrame(
        index=['INS', 'GCG', 'PPY'],
        columns=['cell1', 'cell2', 'cell3'],
        data=[
            [2302, 123, 0],
            [0, 5034, 6453],
            [0, 0, 1]],
        )
    n_pcs = 2
    k = 1
    threshold = 0.8

    sa = Averages(
            'Enge_2017',
            n_neighbors=k,
            threshold_neighborhood=threshold,
            n_pcs=n_pcs,
            distance_metric='correlation',
            n_cells_per_type=20,
            n_neighbors_out_of_atlas=1
            )

    sa.new_data = matrix
    sa._check_init_arguments()
    sa.fetch_atlas_if_needed()
    sa.compute_feature_intersection()
    sa._check_feature_intersection()
    sa.prepare_feature_selection()
    sa.select_features()
    sa._check_feature_selection()
    sa.merge_atlas_newdata()
    sa.compute_pca()
    sa.compute_similarity_graph()
    neis = sa.neighbors

    assert(isinstance(neis, list))
    assert(len(neis) == int(np.sum(sa.sizes)))
    for nei in neis:
        assert(isinstance(nei, list))


def test_run_mock_cells():
    aname = 'Darmanis_2015_nofetal'
    atlas = AtlasFetcher().fetch_atlas(
            aname, kind='average')

    ind = [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 4, 5]
    matrix = atlas[ind]
    cell_types = atlas.obs_names.values[ind]

    sa = Averages(
            aname,
            n_pcs=10,
            )
    sa.fit(matrix)

    assert((cell_types == sa.membership).mean() > 0.9)


def test_run_within_atlas():
    aname = 'Darmanis_2015_nofetal'
    atlas = AtlasFetcher().fetch_atlas(
            aname, kind='subsample')

    ind = [
        0, 2, 5, 8, 10, 15, 20, 25, 28, 30, 35, 38,
        40, 45, 50, 60, 70, 75, 80, 90]
    matrix = atlas[ind]
    cell_types = atlas.obs['CellType'].values[ind]

    sa = Averages(
            aname,
            n_features_per_cell_type=2,
            n_features_overdispersed=5,
            n_pcs=9,
            )
    sa.fit(matrix)

    for i in range(len(cell_types)):
        print('{:10s}    {:10s}'.format(cell_types[i], sa.membership[i]))
    print((cell_types == sa.membership).mean())
    assert((cell_types == sa.membership).mean() >= 0.5)


if __name__ == '__main__':

    test_run_within_atlas()
